@@ -14,6 +14,7 @@
 #    under the License.
 
 import gzip
+import io
 import logging
 import math
 import os
@@ -31,7 +32,7 @@ from oslo_utils import excutils
 from oslo_utils import imageutils
 from oslo_utils import units
 import requests
-import six
+import tenacity
 
 from ironic_lib.common.i18n import _
 from ironic_lib import disk_partitioner
@@ -51,23 +52,21 @@ opts = [
     cfg.StrOpt('dd_block_size',
                default='1M',
                help='Block size to use when writing to the nodes disk.'),
-    # TODO(dtantsur): remove in Ussuri, then clean up ironic from it.
-    cfg.IntOpt('iscsi_verify_attempts',
-               default=3,
-               help='Maximum attempts to verify an iSCSI connection is '
-                    'active, sleeping 1 second between attempts.',
-               deprecated_for_removal=True,
-               deprecated_reason='Split into [iscsi]verify_attempts and '
-                                 'partition_detection_attempts.'),
     cfg.IntOpt('partition_detection_attempts',
+               default=3,
                min=1,
-               help='Maximum attempts to detect a newly created partition. '
-                    'Defaults to iscsi_verify_attempts, will default to 3 '
-                    'when it is removed.'),
+               help='Maximum attempts to detect a newly created partition.'),
     cfg.IntOpt('partprobe_attempts',
                default=10,
                help='Maximum number of attempts to try to read the '
                     'partition.'),
+    cfg.IntOpt('image_convert_memory_limit',
+               default=1024,
+               help='Memory limit for "qemu-img convert" in MiB. Implemented '
+                    'via the address space resource limit.'),
+    cfg.IntOpt('image_convert_attempts',
+               default=3,
+               help='Number of attempts to convert an image.'),
 ]
 
 CONF = cfg.CONF
@@ -76,7 +75,7 @@ CONF.register_opts(opts, group='disk_utils')
 LOG = logging.getLogger(__name__)
 
 _PARTED_PRINT_RE = re.compile(r"^(\d+):([\d\.]+)MiB:"
-                              "([\d\.]+)MiB:([\d\.]+)MiB:(\w*):(.*):(.*);")
+                              r"([\d\.]+)MiB:([\d\.]+)MiB:(\w*):(.*):(.*);")
 
 CONFIGDRIVE_LABEL = "config-2"
 MAX_CONFIG_DRIVE_SIZE_MB = 64
@@ -85,7 +84,16 @@ MAX_CONFIG_DRIVE_SIZE_MB = 64
 MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR = 2097152
 
 # Limit the memory address space to 1 GiB when running qemu-img
-QEMU_IMG_LIMITS = processutils.ProcessLimits(address_space=1 * units.Gi)
+QEMU_IMG_LIMITS = None
+
+
+def _qemu_img_limits():
+    global QEMU_IMG_LIMITS
+    if QEMU_IMG_LIMITS is None:
+        QEMU_IMG_LIMITS = processutils.ProcessLimits(
+            address_space=CONF.disk_utils.image_convert_memory_limit
+            * units.Mi)
+    return QEMU_IMG_LIMITS
 
 
 def list_partitions(device):
@@ -291,8 +299,8 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
     # a PrEP partition at the start of the disk. This is an 8 MiB partition
     # with the boot and prep flags set. The bootloader should be installed
     # here.
-    if (cpu_arch.startswith("ppc64") and boot_mode == "bios" and
-        boot_option == "local"):
+    if (cpu_arch.startswith("ppc64") and boot_mode == "bios"
+            and boot_option == "local"):
         LOG.debug("Add PReP boot partition (8 MB) to device: "
                   "%(dev)s for node %(node)s",
                   {'dev': dev, 'node': node_uuid})
@@ -338,15 +346,13 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
     if commit:
         # write to the disk
         dp.commit()
-        _trigger_device_rescan(dev)
+        trigger_device_rescan(dev)
     return part_dict
 
 
 def is_block_device(dev):
     """Check whether a device is block or not."""
-    attempts = (CONF.disk_utils.iscsi_verify_attempts
-                if CONF.disk_utils.partition_detection_attempts is None
-                else CONF.disk_utils.partition_detection_attempts)
+    attempts = CONF.disk_utils.partition_detection_attempts
     for attempt in range(attempts):
         try:
             s = os.stat(dev)
@@ -382,14 +388,34 @@ def qemu_img_info(path):
 
     out, err = utils.execute('env', 'LC_ALL=C', 'LANG=C',
                              'qemu-img', 'info', path,
-                             prlimit=QEMU_IMG_LIMITS)
+                             prlimit=_qemu_img_limits())
     return imageutils.QemuImgInfo(out)
 
 
+def _retry_on_res_temp_unavailable(exc):
+    if (isinstance(exc, processutils.ProcessExecutionError)
+            and 'Resource temporarily unavailable' in exc.stderr):
+        return True
+    return False
+
+
+@tenacity.retry(
+    retry=tenacity.retry_if_exception(_retry_on_res_temp_unavailable),
+    stop=tenacity.stop_after_attempt(CONF.disk_utils.image_convert_attempts),
+    reraise=True)
 def convert_image(source, dest, out_format, run_as_root=False):
     """Convert image to other format."""
     cmd = ('qemu-img', 'convert', '-O', out_format, source, dest)
-    utils.execute(*cmd, run_as_root=run_as_root, prlimit=QEMU_IMG_LIMITS)
+    try:
+        utils.execute(*cmd, run_as_root=run_as_root,
+                      prlimit=_qemu_img_limits(),
+                      use_standard_locale=True)
+    except processutils.ProcessExecutionError as e:
+        if 'Resource temporarily unavailable' in e.stderr:
+            LOG.debug('Failed to convert image, retrying. Error: %s', e)
+            # Sync disk caches before the next attempt
+            utils.execute('sync')
+        raise
 
 
 def populate_image(src, dst, conv_flags=None):
@@ -519,7 +545,7 @@ def _get_configdrive(configdrive, node_uuid, tempdir=None):
         data = configdrive
 
     try:
-        data = six.BytesIO(base64.decode_as_bytes(data))
+        data = io.BytesIO(base64.decode_as_bytes(data))
     except TypeError:
         error_msg = (_('Config drive for node %s is not base64 encoded '
                        'or the content is malformed.') % node_uuid)
@@ -747,11 +773,8 @@ def _get_labelled_partition(device_path, label, node_uuid):
     :returns: block device file for partition if it exists; otherwise it
               returns None.
     """
+    partprobe(device_path)
     try:
-        utils.execute('partprobe', device_path, run_as_root=True,
-                      attempts=CONF.disk_utils.partprobe_attempts)
-
-        # lsblk command
         output, err = utils.execute('lsblk', '-Po', 'name,label', device_path,
                                     check_exit_code=[0, 1],
                                     use_standard_locale=True, run_as_root=True)
@@ -856,7 +879,50 @@ def fix_gpt_partition(device, node_uuid):
         raise exception.InstanceDeployFailure(msg)
 
 
-def _trigger_device_rescan(device):
+def udev_settle():
+    """Wait for the udev event queue to settle.
+
+    Wait for the udev event queue to settle to make sure all devices
+    are detected once the machine boots up.
+
+    :return: True on success, False otherwise.
+    """
+    LOG.debug('Waiting until udev event queue is empty')
+    try:
+        utils.execute('udevadm', 'settle')
+    except processutils.ProcessExecutionError as e:
+        LOG.warning('Something went wrong when waiting for udev '
+                    'to settle. Error: %s', e)
+        return False
+    else:
+        return True
+
+
+def partprobe(device, attempts=None):
+    """Probe partitions on the given device.
+
+    :param device: The block device containing paritions that is attempting
+                   to be updated.
+    :param attempts: Number of attempts to run partprobe, the default is read
+                     from the configuration.
+    :return: True on success, False otherwise.
+    """
+    if attempts is None:
+        attempts = CONF.disk_utils.partprobe_attempts
+
+    try:
+        utils.execute('partprobe', device, run_as_root=True, attempts=attempts)
+    except (processutils.UnknownArgumentError,
+            processutils.ProcessExecutionError, OSError) as e:
+        LOG.warning("Unable to probe for partitions on device %(device)s, "
+                    "the partitioning table may be broken. Error: %(error)s",
+                    {'device': device, 'error': e})
+        return False
+    else:
+        return True
+
+
+def trigger_device_rescan(device, attempts=None):
     """Sync and trigger device rescan.
 
     Disk partition performed via parted, when performed on a ramdisk
@@ -874,25 +940,25 @@ def _trigger_device_rescan(device):
 
     :param device: The block device containing paritions that is attempting
                    to be updated.
+    :param attempts: Number of attempts to run partprobe, the default is read
+                     from the configuration.
+    :return: True on success, False otherwise.
     """
-    # TODO(TheJulia): This helper method was broken out for re-use, and
-    # does not have an explicit test case for itself, and it should, but
-    # it's steps are fairly explicitly checked with mocks on execute in
-    # the partition code.
-    LOG.debug('Explicitly calling sync to force buffer/cache flush.')
+    LOG.debug('Explicitly calling sync to force buffer/cache flush')
     utils.execute('sync')
     # Make sure any additions to the partitioning are reflected in the
     # kernel.
-    LOG.debug('Waiting until udev event queue is empty')
-    utils.execute('udevadm', 'settle')
+    udev_settle()
+    partprobe(device, attempts=attempts)
     try:
-        utils.execute('partprobe', device, run_as_root=True,
-                      attempts=CONF.disk_utils.partprobe_attempts)
         # Also verify that the partitioning is correct now.
         utils.execute('sgdisk', '-v', device, run_as_root=True)
     except processutils.ProcessExecutionError as exc:
-        LOG.warning('Failed to verify partitioning after creating '
-                    'partition(s): %s', exc)
+        LOG.warning('Failed to verify partition tables on device %(dev)s: '
+                    '%(err)s', {'dev': device, 'err': exc})
+        return False
+    else:
+        return True
 
 
 def create_config_drive_partition(node_uuid, device, configdrive):
@@ -917,11 +983,11 @@ def create_config_drive_partition(node_uuid, device, configdrive):
         confdrive_mb, confdrive_file = _get_configdrive(configdrive,
                                                         node_uuid)
         if confdrive_mb > MAX_CONFIG_DRIVE_SIZE_MB:
-                raise exception.InstanceDeployFailure(
-                    _('Config drive size exceeds maximum limit of 64MiB. '
-                      'Size of the given config drive is %(size)d MiB for '
-                      'node %(node)s.')
-                    % {'size': confdrive_mb, 'node': node_uuid})
+            raise exception.InstanceDeployFailure(
+                _('Config drive size exceeds maximum limit of 64MiB. '
+                  'Size of the given config drive is %(size)d MiB for '
+                  'node %(node)s.')
+                % {'size': confdrive_mb, 'node': node_uuid})
 
         LOG.debug("Adding config drive partition %(size)d MiB to "
                   "device: %(dev)s for node %(node)s",
@@ -970,15 +1036,15 @@ def create_config_drive_partition(node_uuid, device, configdrive):
                                 "node %(node)s. Creating config drive "
                                 "at the end of the disk %(disk)s.",
                                 {'node': node_uuid, 'disk': device})
-                    startlimit = (MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR -
-                                  MAX_CONFIG_DRIVE_SIZE_MB - 1)
+                    startlimit = (MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR
+                                  - MAX_CONFIG_DRIVE_SIZE_MB - 1)
                     endlimit = MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR - 1
 
                 utils.execute('parted', '-a', 'optimal', '-s', '--', device,
                               'mkpart', 'primary', 'fat32', startlimit,
                               endlimit, run_as_root=True)
             # Trigger device rescan
-            _trigger_device_rescan(device)
+            trigger_device_rescan(device)
 
             upd_parts = set(part['number'] for part in list_partitions(device))
             new_part = set(upd_parts) - set(cur_parts)
@@ -995,8 +1061,7 @@ def create_config_drive_partition(node_uuid, device, configdrive):
             else:
                 config_drive_part = '%s%s' % (device, new_part.pop())
 
-            LOG.debug('Waiting until udev event queue is empty')
-            utils.execute('udevadm', 'settle')
+            udev_settle()
 
             # NOTE(vsaienko): check that devise actually exists,
             # it is not handled by udevadm when using ISCSI, for more info see:
