@@ -18,8 +18,8 @@ import os
 import shutil
 import stat
 import tempfile
+from unittest import mock
 
-import mock
 from oslo_concurrency import processutils
 from oslo_config import cfg
 from oslo_serialization import base64
@@ -47,9 +47,9 @@ BYT;
 """
         expected = [
             {'number': 1, 'start': 1, 'end': 501, 'size': 500,
-             'filesystem': 'ext4', 'flags': 'boot'},
+             'filesystem': 'ext4', 'partition_name': '', 'flags': 'boot'},
             {'number': 2, 'start': 501, 'end': 476940, 'size': 476439,
-             'filesystem': '', 'flags': ''},
+             'filesystem': '', 'partition_name': '', 'flags': ''},
         ]
         execute_mock.return_value = (output, '')
         result = disk_utils.list_partitions('/dev/fake')
@@ -70,6 +70,46 @@ BYT;
         self.assertEqual(1, log_mock.call_count)
 
 
+@mock.patch.object(utils, 'execute', autospec=True)
+class ListPartitionsGPTTestCase(base.IronicLibTestCase):
+
+    def test_correct(self, execute_mock):
+        output = """
+BYT;
+/dev/vda:40960MiB:virtblk:512:512:gpt:Virtio Block Device:;
+2:1.00MiB:2.00MiB:1.00MiB::Bios partition:bios_grub;
+1:4.00MiB:5407MiB:5403MiB:ext4:Root partition:;
+3:5407MiB:5507MiB:100MiB:fat16:Boot partition:boot, esp;
+"""
+        expected = [
+            {'end': 2, 'number': 2, 'start': 1, 'flags': 'bios_grub',
+             'filesystem': '', 'partition_name': 'Bios partition', 'size': 1},
+            {'end': 5407, 'number': 1, 'start': 4, 'flags': '',
+             'filesystem': 'ext4', 'partition_name': 'Root partition',
+             'size': 5403},
+            {'end': 5507, 'number': 3, 'start': 5407,
+             'flags': 'boot, esp', 'filesystem': 'fat16',
+             'partition_name': 'Boot partition', 'size': 100},
+        ]
+        execute_mock.return_value = (output, '')
+        result = disk_utils.list_partitions('/dev/fake')
+        self.assertEqual(expected, result)
+        execute_mock.assert_called_once_with(
+            'parted', '-s', '-m', '/dev/fake', 'unit', 'MiB', 'print',
+            use_standard_locale=True, run_as_root=True)
+
+    @mock.patch.object(disk_utils.LOG, 'warning', autospec=True)
+    def test_incorrect(self, log_mock, execute_mock):
+        output = """
+BYT;
+/dev/vda:40960MiB:virtblk:512:512:gpt:Virtio Block Device:;
+2:XX1.00MiB:---:1.00MiB::primary:bios_grub;
+"""
+        execute_mock.return_value = (output, '')
+        self.assertEqual([], disk_utils.list_partitions('/dev/fake'))
+        self.assertEqual(1, log_mock.call_count)
+
+
 @mock.patch.object(disk_partitioner.DiskPartitioner, 'commit', lambda _: None)
 class WorkOnDiskTestCase(base.IronicLibTestCase):
 
@@ -202,7 +242,42 @@ class WorkOnDiskTestCase(base.IronicLibTestCase):
 
     @mock.patch.object(utils, 'mkfs', lambda fs, path, label=None: None)
     @mock.patch.object(disk_utils, 'block_uuid', lambda p: 'uuid')
-    @mock.patch.object(disk_utils, 'populate_image', lambda *_: None)
+    @mock.patch.object(disk_utils, 'populate_image', autospec=True)
+    def test_without_image(self, mock_populate):
+        ephemeral_part = '/dev/fake-part1'
+        swap_part = '/dev/fake-part2'
+        root_part = '/dev/fake-part3'
+        ephemeral_mb = 256
+        ephemeral_format = 'exttest'
+
+        self.mock_mp.return_value = {'ephemeral': ephemeral_part,
+                                     'swap': swap_part,
+                                     'root': root_part}
+        self.mock_ibd.return_value = True
+        calls = [mock.call(root_part),
+                 mock.call(swap_part),
+                 mock.call(ephemeral_part)]
+        res = disk_utils.work_on_disk(self.dev, self.root_mb,
+                                      self.swap_mb, ephemeral_mb,
+                                      ephemeral_format,
+                                      None, self.node_uuid)
+        self.assertEqual(self.mock_ibd.call_args_list, calls)
+        self.mock_mp.assert_called_once_with(self.dev, self.root_mb,
+                                             self.swap_mb, ephemeral_mb,
+                                             self.configdrive_mb,
+                                             self.node_uuid, commit=True,
+                                             boot_option="netboot",
+                                             boot_mode="bios",
+                                             disk_label=None,
+                                             cpu_arch="")
+        self.assertEqual(root_part, res['partitions']['root'])
+        self.assertEqual('uuid', res['root uuid'])
+        self.assertFalse(mock_populate.called)
+
+    @mock.patch.object(utils, 'mkfs', lambda fs, path, label=None: None)
+    @mock.patch.object(disk_utils, 'block_uuid', lambda p: 'uuid')
+    @mock.patch.object(disk_utils, 'populate_image', lambda image_path,
+                       root_path, conv_flags=None: None)
     def test_gpt_disk_label(self):
         ephemeral_part = '/dev/fake-part1'
         swap_part = '/dev/fake-part2'
@@ -220,7 +295,7 @@ class WorkOnDiskTestCase(base.IronicLibTestCase):
         disk_utils.work_on_disk(self.dev, self.root_mb,
                                 self.swap_mb, ephemeral_mb, ephemeral_format,
                                 self.image_path, self.node_uuid,
-                                disk_label='gpt')
+                                disk_label='gpt', conv_flags=None)
         self.assertEqual(self.mock_ibd.call_args_list, calls)
         self.mock_mp.assert_called_once_with(self.dev, self.root_mb,
                                              self.swap_mb, ephemeral_mb,
@@ -263,7 +338,7 @@ class WorkOnDiskTestCase(base.IronicLibTestCase):
         mock_mkfs.assert_called_once_with(fs='vfat', path=efi_part,
                                           label='efi-part')
         mock_populate_image.assert_called_once_with(self.image_path,
-                                                    root_part)
+                                                    root_part, conv_flags=None)
         mock_block_uuid.assert_any_call(root_part)
         mock_block_uuid.assert_any_call(efi_part)
 
@@ -309,7 +384,7 @@ class WorkOnDiskTestCase(base.IronicLibTestCase):
 
         self.mock_mp.return_value = {'PReP Boot partition': prep_part,
                                      'root': root_part}
-        self.mock_ibd.return_value = True
+        self.mock_ibd.return_vaue = True
         calls = [mock.call(root_part),
                  mock.call(prep_part)]
         disk_utils.work_on_disk(self.dev, self.root_mb,
@@ -328,6 +403,92 @@ class WorkOnDiskTestCase(base.IronicLibTestCase):
                                              cpu_arch="ppc64le")
         self.assertFalse(mock_mkfs.called)
 
+    @mock.patch.object(disk_utils, 'block_uuid', autospec=True)
+    @mock.patch.object(disk_utils, 'populate_image', autospec=True)
+    @mock.patch.object(utils, 'mkfs', autospec=True)
+    def test_convert_to_sparse(self, mock_mkfs, mock_populate_image,
+                               mock_block_uuid):
+        ephemeral_part = '/dev/fake-part1'
+        swap_part = '/dev/fake-part2'
+        root_part = '/dev/fake-part3'
+        ephemeral_mb = 256
+        ephemeral_format = 'exttest'
+
+        self.mock_mp.return_value = {'ephemeral': ephemeral_part,
+                                     'swap': swap_part,
+                                     'root': root_part}
+        self.mock_ibd.return_value = True
+        disk_utils.work_on_disk(self.dev, self.root_mb,
+                                self.swap_mb, ephemeral_mb, ephemeral_format,
+                                self.image_path, self.node_uuid,
+                                disk_label='gpt', conv_flags='sparse')
+
+        mock_populate_image.assert_called_once_with(self.image_path,
+                                                    root_part,
+                                                    conv_flags='sparse')
+
+
+class GetUEFIDiskIdentifierTestCase(base.IronicLibTestCase):
+
+    def setUp(self):
+        super(GetUEFIDiskIdentifierTestCase, self).setUp()
+        self.dev = '/dev/fake'
+
+    @mock.patch.object(utils, 'execute', autospec=True)
+    def test_get_uefi_disk_identifier_uefi_bootable_image(self, mock_execute):
+        mock_execute.return_value = ('', '')
+        fdisk_output = """
+Disk /dev/sda: 931.5 GiB, 1000171331584 bytes, 1953459632 sectors
+Units: sectors of 1 * 512 = 512 bytes
+Sector size (logical/physical): 512 bytes / 512 bytes
+I/O size (minimum/optimal): 262144 bytes / 262144 bytes
+Disklabel type: gpt
+Disk identifier: 73457A6C-3595-4965-8D83-2EA1BD85F327
+
+Device          Start        End    Sectors   Size Type
+/dev/fake-part1        2048    1050623    1048576   512M EFI System
+/dev/fake-part2     1050624 1920172031 1919121408 915.1G Linux filesystem
+/dev/fake-part3  1920172032 1953458175   33286144  15.9G Linux swap
+"""
+        partition_id = '/dev/fake-part1'
+        lsblk_output = 'UUID="ABCD-B05B"\n'
+        part_result = 'ABCD-B05B'
+        mock_execute.side_effect = [(fdisk_output, ''), (lsblk_output, '')]
+        result = disk_utils.get_uefi_disk_identifier(self.dev)
+        self.assertEqual(part_result, result)
+        execute_calls = [
+            mock.call('fdisk', '-l', self.dev, run_as_root=True),
+            mock.call('lsblk', '-PbioUUID', partition_id,
+                      run_as_root=True)
+        ]
+        mock_execute.assert_has_calls(execute_calls)
+
+    @mock.patch.object(utils, 'execute', autospec=True)
+    def test_get_uefi_disk_identifier_non_uefi_bootable_image(self,
+                                                              mock_execute):
+        mock_execute.return_value = ('', '')
+        fdisk_output = """
+Disk /dev/vda: 50 GiB, 53687091200 bytes, 104857600 sectors
+Units: sectors of 1 * 512 = 512 bytes
+Sector size (logical/physical): 512 bytes / 512 bytes
+I/O size (minimum/optimal): 512 bytes / 512 bytes
+Disklabel type: dos
+Disk identifier: 0xb82b9faf
+
+Device     Boot Start       End   Sectors Size Id Type
+/dev/fake-part1  *     2048 104857566 104855519  50G 83 Linux
+"""
+        partition_id = None
+        mock_execute.side_effect = [(fdisk_output, ''),
+                                    processutils.ProcessExecutionError()]
+        self.assertRaises(exception.InstanceDeployFailure,
+                          disk_utils.get_uefi_disk_identifier, self.dev)
+        execute_calls = [
+            mock.call('fdisk', '-l', self.dev, run_as_root=True),
+            mock.call('lsblk', '-PbioUUID', partition_id, run_as_root=True)
+        ]
+        mock_execute.assert_has_calls(execute_calls)
+
 
 @mock.patch.object(utils, 'execute', autospec=True)
 class MakePartitionsTestCase(base.IronicLibTestCase):
@@ -350,6 +511,12 @@ class MakePartitionsTestCase(base.IronicLibTestCase):
         return ['parted', '-a', 'optimal', '-s', dev,
                 '--', 'unit', 'MiB', 'mklabel', label]
 
+    def _add_efi_sz(self, x):
+        return str(x + self.efi_size)
+
+    def _add_bios_sz(self, x):
+        return str(x + self.bios_size)
+
     def _test_make_partitions(self, mock_exc, boot_option, boot_mode='bios',
                               disk_label=None, cpu_arch=""):
         mock_exc.return_value = ('', '')
@@ -360,14 +527,13 @@ class MakePartitionsTestCase(base.IronicLibTestCase):
                                    cpu_arch=cpu_arch)
 
         if boot_option == "local" and boot_mode == "uefi":
-            add_efi_sz = lambda x: str(x + self.efi_size)
             expected_mkpart = ['mkpart', 'primary', 'fat32', '1',
-                               add_efi_sz(1),
+                               self._add_efi_sz(1),
                                'set', '1', 'boot', 'on',
                                'mkpart', 'primary', 'linux-swap',
-                               add_efi_sz(1), add_efi_sz(513), 'mkpart',
-                               'primary', '', add_efi_sz(513),
-                               add_efi_sz(1537)]
+                               self._add_efi_sz(1), self._add_efi_sz(513),
+                               'mkpart', 'primary', '', self._add_efi_sz(513),
+                               self._add_efi_sz(1537)]
         else:
             if boot_option == "local":
                 if disk_label == "gpt":
@@ -378,14 +544,15 @@ class MakePartitionsTestCase(base.IronicLibTestCase):
                                            '9', '521', 'mkpart', 'primary',
                                            '', '521', '1545']
                     else:
-                        add_bios_sz = lambda x: str(x + self.bios_size)
                         expected_mkpart = ['mkpart', 'primary', '', '1',
-                                           add_bios_sz(1),
+                                           self._add_bios_sz(1),
                                            'set', '1', 'bios_grub', 'on',
                                            'mkpart', 'primary', 'linux-swap',
-                                           add_bios_sz(1), add_bios_sz(513),
+                                           self._add_bios_sz(1),
+                                           self._add_bios_sz(513),
                                            'mkpart', 'primary', '',
-                                           add_bios_sz(513), add_bios_sz(1537)]
+                                           self._add_bios_sz(513),
+                                           self._add_bios_sz(1537)]
                 elif cpu_arch.startswith('ppc64'):
                     expected_mkpart = ['mkpart', 'primary', '', '1', '9',
                                        'set', '1', 'boot', 'on',
@@ -402,14 +569,21 @@ class MakePartitionsTestCase(base.IronicLibTestCase):
                                    '513', 'mkpart', 'primary', '', '513',
                                    '1537']
         self.dev = 'fake-dev'
-        parted_cmd = (self._get_parted_cmd(self.dev, disk_label) +
-                      expected_mkpart)
+        parted_cmd = (self._get_parted_cmd(self.dev, disk_label)
+                      + expected_mkpart)
         parted_call = mock.call(*parted_cmd, use_standard_locale=True,
                                 run_as_root=True, check_exit_code=[0])
         fuser_cmd = ['fuser', 'fake-dev']
         fuser_call = mock.call(*fuser_cmd, run_as_root=True,
                                check_exit_code=[0, 1])
-        mock_exc.assert_has_calls([parted_call, fuser_call])
+
+        sync_calls = [mock.call('sync'),
+                      mock.call('udevadm', 'settle'),
+                      mock.call('partprobe', self.dev, attempts=10,
+                                run_as_root=True),
+                      mock.call('sgdisk', '-v', self.dev, run_as_root=True)]
+
+        mock_exc.assert_has_calls([parted_call, fuser_call] + sync_calls)
 
     def test_make_partitions(self, mock_exc):
         self._test_make_partitions(mock_exc, boot_option="netboot")
@@ -522,7 +696,7 @@ class MakePartitionsTestCase(base.IronicLibTestCase):
         self.assertEqual(expected_result, result)
 
 
-@mock.patch.object(utils, 'execute', autospec=True, return_value=('', ''))
+@mock.patch.object(utils, 'execute', autospec=True)
 class DestroyMetaDataTestCase(base.IronicLibTestCase):
 
     def setUp(self):
@@ -531,9 +705,31 @@ class DestroyMetaDataTestCase(base.IronicLibTestCase):
         self.node_uuid = "12345678-1234-1234-1234-1234567890abcxyz"
 
     def test_destroy_disk_metadata(self, mock_exec):
+        # Note(TheJulia): This list will get-reused, but only the second
+        # execution returning a string is needed for the test as otherwise
+        # command output is not used.
+        mock_exec.side_effect = iter([
+            (None, None),
+            ('1024\n', None),
+            (None, None),
+            (None, None),
+            (None, None),
+            (None, None)])
+
         expected_calls = [mock.call('wipefs', '--force', '--all', 'fake-dev',
                                     run_as_root=True,
                                     use_standard_locale=True),
+                          mock.call('blockdev', '--getsz', 'fake-dev',
+                                    check_exit_code=[0],
+                                    run_as_root=True),
+                          mock.call('dd', 'bs=512', 'if=/dev/zero',
+                                    'of=fake-dev', 'count=33',
+                                    run_as_root=True,
+                                    use_standard_locale=True),
+                          mock.call('dd', 'bs=512', 'if=/dev/zero',
+                                    'of=fake-dev', 'count=33', 'seek=991',
+                                    run_as_root=True,
+                                    use_standard_locale=True),
                           mock.call('sgdisk', '-Z', 'fake-dev',
                                     run_as_root=True,
                                     use_standard_locale=True),
@@ -559,11 +755,26 @@ class DestroyMetaDataTestCase(base.IronicLibTestCase):
         expected_calls = [mock.call('wipefs', '--force', '--all', 'fake-dev',
                                     run_as_root=True,
                                     use_standard_locale=True),
+                          mock.call('blockdev', '--getsz', 'fake-dev',
+                                    check_exit_code=[0],
+                                    run_as_root=True),
+                          mock.call('dd', 'bs=512', 'if=/dev/zero',
+                                    'of=fake-dev', 'count=33',
+                                    run_as_root=True,
+                                    use_standard_locale=True),
+                          mock.call('dd', 'bs=512', 'if=/dev/zero',
+                                    'of=fake-dev', 'count=33', 'seek=991',
+                                    run_as_root=True,
+                                    use_standard_locale=True),
                           mock.call('sgdisk', '-Z', 'fake-dev',
                                     run_as_root=True,
                                     use_standard_locale=True)]
-        mock_exec.side_effect = [(None, None),
-                                 processutils.ProcessExecutionError()]
+        mock_exec.side_effect = iter([
+            (None, None),
+            ('1024\n', None),
+            (None, None),
+            (None, None),
+            processutils.ProcessExecutionError()])
         self.assertRaises(processutils.ProcessExecutionError,
                           disk_utils.destroy_disk_metadata,
                           self.dev,
@@ -571,11 +782,14 @@ class DestroyMetaDataTestCase(base.IronicLibTestCase):
         mock_exec.assert_has_calls(expected_calls)
 
     def test_destroy_disk_metadata_wipefs_not_support_force(self, mock_exec):
-        mock_exec.side_effect = iter(
-            [processutils.ProcessExecutionError(description='--force'),
-             (None, None),
-             (None, None),
-             ('', '')])
+        mock_exec.side_effect = iter([
+            processutils.ProcessExecutionError(description='--force'),
+            (None, None),
+            ('1024\n', None),
+            (None, None),
+            (None, None),
+            (None, None),
+            (None, None)])
 
         expected_call = [mock.call('wipefs', '--force', '--all', 'fake-dev',
                                    run_as_root=True,
@@ -612,7 +826,15 @@ class PopulateImageTestCase(base.IronicLibTestCase):
         type(mock_qinfo.return_value).file_format = mock.PropertyMock(
             return_value='raw')
         disk_utils.populate_image('src', 'dst')
-        mock_dd.assert_called_once_with('src', 'dst')
+        mock_dd.assert_called_once_with('src', 'dst', conv_flags=None)
+        self.assertFalse(mock_cg.called)
+
+    def test_populate_raw_image_with_convert(self, mock_cg, mock_qinfo,
+                                             mock_dd):
+        type(mock_qinfo.return_value).file_format = mock.PropertyMock(
+            return_value='raw')
+        disk_utils.populate_image('src', 'dst', conv_flags='sparse')
+        mock_dd.assert_called_once_with('src', 'dst', conv_flags='sparse')
         self.assertFalse(mock_cg.called)
 
     def test_populate_qcow2_image(self, mock_cg, mock_qinfo, mock_dd):
@@ -623,6 +845,9 @@ class PopulateImageTestCase(base.IronicLibTestCase):
         self.assertFalse(mock_dd.called)
 
 
+# NOTE(TheJulia): trigger_device_rescan is systemwide thus pointless
+# to execute in the file test case. Also, CI unit test jobs lack sgdisk.
+@mock.patch.object(disk_utils, 'trigger_device_rescan', lambda *_: None)
 @mock.patch.object(utils, 'wait_for_disk_to_become_available', lambda *_: None)
 @mock.patch.object(disk_utils, 'is_block_device', lambda d: True)
 @mock.patch.object(disk_utils, 'block_uuid', lambda p: 'uuid')
@@ -780,6 +1005,16 @@ class OtherFunctionTestCase(base.IronicLibTestCase):
                           disk_utils.is_block_device, device)
         mock_os.assert_has_calls([mock.call(device)] * 3)
 
+    @mock.patch.object(os, 'stat', autospec=True)
+    def test_is_block_device_attempts(self, mock_os):
+        CONF.set_override('partition_detection_attempts', 2,
+                          group='disk_utils')
+        device = '/dev/disk/by-path/ip-1.2.3.4:5678-iscsi-iqn.fake-lun-9'
+        mock_os.side_effect = OSError
+        self.assertRaises(exception.InstanceDeployFailure,
+                          disk_utils.is_block_device, device)
+        mock_os.assert_has_calls([mock.call(device)] * 2)
+
     @mock.patch.object(imageutils, 'QemuImgInfo', autospec=True)
     @mock.patch.object(os.path, 'exists', return_value=False, autospec=True)
     def test_qemu_img_info_path_doesnt_exist(self, path_exists_mock,
@@ -807,7 +1042,76 @@ class OtherFunctionTestCase(base.IronicLibTestCase):
         execute_mock.assert_called_once_with('qemu-img', 'convert', '-O',
                                              'out_format', 'source', 'dest',
                                              run_as_root=False,
-                                             prlimit=mock.ANY)
+                                             prlimit=mock.ANY,
+                                             use_standard_locale=True)
+
+    @mock.patch.object(utils, 'execute', autospec=True)
+    def test_convert_image_retries(self, execute_mock):
+        ret_err = 'qemu: qemu_thread_create: Resource temporarily unavailable'
+        execute_mock.side_effect = [
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            ('', ''),
+        ]
+
+        disk_utils.convert_image('source', 'dest', 'out_format')
+        convert_call = mock.call('qemu-img', 'convert', '-O',
+                                 'out_format', 'source', 'dest',
+                                 run_as_root=False,
+                                 prlimit=mock.ANY,
+                                 use_standard_locale=True)
+        execute_mock.assert_has_calls([
+            convert_call,
+            mock.call('sync'),
+            convert_call,
+            mock.call('sync'),
+            convert_call,
+        ])
+
+    @mock.patch.object(utils, 'execute', autospec=True)
+    def test_convert_image_retries_and_fails(self, execute_mock):
+        ret_err = 'qemu: qemu_thread_create: Resource temporarily unavailable'
+        execute_mock.side_effect = [
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err), ('', ''),
+            processutils.ProcessExecutionError(stderr=ret_err),
+        ]
+
+        self.assertRaises(processutils.ProcessExecutionError,
+                          disk_utils.convert_image,
+                          'source', 'dest', 'out_format')
+        convert_call = mock.call('qemu-img', 'convert', '-O',
+                                 'out_format', 'source', 'dest',
+                                 run_as_root=False,
+                                 prlimit=mock.ANY,
+                                 use_standard_locale=True)
+        execute_mock.assert_has_calls([
+            convert_call,
+            mock.call('sync'),
+            convert_call,
+            mock.call('sync'),
+            convert_call,
+        ])
+
+    @mock.patch.object(utils, 'execute', autospec=True)
+    def test_convert_image_just_fails(self, execute_mock):
+        ret_err = 'Aliens'
+        execute_mock.side_effect = [
+            processutils.ProcessExecutionError(stderr=ret_err),
+        ]
+
+        self.assertRaises(processutils.ProcessExecutionError,
+                          disk_utils.convert_image,
+                          'source', 'dest', 'out_format')
+        convert_call = mock.call('qemu-img', 'convert', '-O',
+                                 'out_format', 'source', 'dest',
+                                 run_as_root=False,
+                                 prlimit=mock.ANY,
+                                 use_standard_locale=True)
+        execute_mock.assert_has_calls([
+            convert_call,
+        ])
 
     @mock.patch.object(os.path, 'getsize', autospec=True)
     @mock.patch.object(disk_utils, 'qemu_img_info', autospec=True)
@@ -909,6 +1213,22 @@ class WholeDiskPartitionTestCases(base.IronicLibTestCase):
         ]
         mock_execute.assert_has_calls(execute_calls)
 
+    def test_get_partition_present_uppercase(self, mock_execute):
+        lsblk_output = 'NAME="fake12" LABEL="CONFIG-2"\n'
+        part_result = '/dev/fake12'
+        mock_execute.side_effect = [(None, ''), (lsblk_output, '')]
+        result = disk_utils._get_labelled_partition(self.dev,
+                                                    self.config_part_label,
+                                                    self.node_uuid)
+        self.assertEqual(part_result, result)
+        execute_calls = [
+            mock.call('partprobe', self.dev, run_as_root=True, attempts=10),
+            mock.call('lsblk', '-Po', 'name,label', self.dev,
+                      check_exit_code=[0, 1],
+                      use_standard_locale=True, run_as_root=True)
+        ]
+        mock_execute.assert_has_calls(execute_calls)
+
     def test_get_partition_absent(self, mock_execute):
         mock_execute.side_effect = [(None, ''),
                                     (None, '')]
@@ -949,8 +1269,13 @@ class WholeDiskPartitionTestCases(base.IronicLibTestCase):
                                'Failed to retrieve partition labels',
                                disk_utils._get_labelled_partition, self.dev,
                                self.config_part_label, self.node_uuid)
-        mock_execute.assert_called_once_with(
-            'partprobe', self.dev, run_as_root=True, attempts=10)
+        execute_calls = [
+            mock.call('partprobe', self.dev, run_as_root=True, attempts=10),
+            mock.call('lsblk', '-Po', 'name,label', self.dev,
+                      check_exit_code=[0, 1],
+                      use_standard_locale=True, run_as_root=True)
+        ]
+        mock_execute.assert_has_calls(execute_calls)
         self.assertEqual(1, mock_log.call_count)
 
     def _test_is_disk_larger_than_max_size(self, mock_execute, blk_out):
@@ -1022,18 +1347,17 @@ class WholeDiskPartitionTestCases(base.IronicLibTestCase):
         self.assertEqual(1, mock_log.call_count)
 
     def test_fix_gpt_structs_fix_required(self, mock_execute):
-        partprobe_err = """
-Error: The backup GPT table is not at the end of the disk, as it should be.
-This might mean that another operating system believes the disk is smaller.
-Fix, by moving the backup to the end (and removing the old backup)?
-Warning: Not all of the space available to /dev/sdb appears to be used,
-you can fix the GPT to use all of the space (an extra 581456476 blocks)
-or continue with the current setting?
+        sgdisk_v_output = """
+Problem: The secondary header's self-pointer indicates that it doesn't reside
+at the end of the disk. If you've added a disk to a RAID array, use the 'e'
+option on the experts' menu to adjust the secondary header's and partition
+table's locations.
+
+Identified 1 problems!
 """
-        mock_execute.return_value = ('', partprobe_err)
+        mock_execute.return_value = (sgdisk_v_output, '')
         execute_calls = [
-            mock.call('partprobe', '/dev/fake', use_standard_locale=True,
-                      run_as_root=True),
+            mock.call('sgdisk', '-v', '/dev/fake', run_as_root=True),
             mock.call('sgdisk', '-e', '/dev/fake', run_as_root=True)
         ]
         disk_utils._fix_gpt_structs('/dev/fake', self.node_uuid)
@@ -1043,8 +1367,7 @@ or continue with the current setting?
         mock_execute.return_value = ('', '')
 
         disk_utils._fix_gpt_structs('/dev/fake', self.node_uuid)
-        mock_execute.assert_called_once_with('partprobe', '/dev/fake',
-                                             use_standard_locale=True,
+        mock_execute.assert_called_once_with('sgdisk', '-v', '/dev/fake',
                                              run_as_root=True)
 
     @mock.patch.object(disk_utils.LOG, 'error', autospec=True)
@@ -1054,8 +1377,7 @@ or continue with the current setting?
                                'Failed to fix GPT data structures on disk',
                                disk_utils._fix_gpt_structs,
                                self.dev, self.node_uuid)
-        mock_execute.assert_called_once_with('partprobe', '/dev/fake',
-                                             use_standard_locale=True,
+        mock_execute.assert_called_once_with('sgdisk', '-v', '/dev/fake',
                                              run_as_root=True)
         self.assertEqual(1, mock_log.call_count)
 
@@ -1166,6 +1488,11 @@ class WholeDiskConfigDriveTestCases(base.IronicLibTestCase):
         mock_execute.assert_has_calls([
             mock.call('sgdisk', '-n', '0:-64MB:0', self.dev,
                       run_as_root=True),
+            mock.call('sync'),
+            mock.call('udevadm', 'settle'),
+            mock.call('partprobe', self.dev, attempts=10, run_as_root=True),
+            mock.call('sgdisk', '-v', self.dev, run_as_root=True),
+
             mock.call('udevadm', 'settle'),
             mock.call('test', '-e', expected_part, attempts=15,
                       check_exit_code=[0], delay_on_retry=True)
@@ -1245,7 +1572,7 @@ class WholeDiskConfigDriveTestCases(base.IronicLibTestCase):
                         self.node_uuid)
             expected_part = '%s-part4' % self.dev
         elif is_nvme_device:
-            self.dev = '/dev/nvmefake'
+            self.dev = '/dev/nvmefake0'
             expected_part = '%sp4' % self.dev
         else:
             expected_part = '/dev/fake4'
@@ -1267,6 +1594,10 @@ class WholeDiskConfigDriveTestCases(base.IronicLibTestCase):
                                     '-0', run_as_root=True)
         mock_execute.assert_has_calls([
             parted_call,
+            mock.call('sync'),
+            mock.call('udevadm', 'settle'),
+            mock.call('partprobe', self.dev, attempts=10, run_as_root=True),
+            mock.call('sgdisk', '-v', self.dev, run_as_root=True),
             mock.call('udevadm', 'settle'),
             mock.call('test', '-e', expected_part, attempts=15,
                       check_exit_code=[0], delay_on_retry=True)
@@ -1363,10 +1694,17 @@ class WholeDiskConfigDriveTestCases(base.IronicLibTestCase):
                                self.node_uuid, self.dev, config_url)
 
         mock_get_configdrive.assert_called_with(config_url, self.node_uuid)
-        mock_execute.assert_called_with('parted', '-a', 'optimal', '-s', '--',
-                                        self.dev, 'mkpart', 'primary',
-                                        'fat32', '-64MiB', '-0',
-                                        run_as_root=True)
+        mock_execute.assert_has_calls([
+            mock.call('parted', '-a', 'optimal', '-s', '--',
+                      self.dev, 'mkpart', 'primary',
+                      'fat32', '-64MiB', '-0',
+                      run_as_root=True),
+            mock.call('sync'),
+            mock.call('udevadm', 'settle'),
+            mock.call('partprobe', self.dev, attempts=10, run_as_root=True),
+            mock.call('sgdisk', '-v', self.dev, run_as_root=True),
+        ])
+
         self.assertEqual(2, mock_list_partitions.call_count)
         mock_fix_gpt_partition.assert_called_with(self.dev, self.node_uuid)
         mock_is_disk_gpt.assert_called_with(self.dev, self.node_uuid)
@@ -1563,3 +1901,37 @@ class WholeDiskConfigDriveTestCases(base.IronicLibTestCase):
         mock_fix_gpt_partition.assert_called_with(self.dev, self.node_uuid)
         mock_is_disk_gpt.assert_called_with(self.dev, self.node_uuid)
         mock_count.assert_called_once_with(self.dev)
+
+
+@mock.patch.object(utils, 'execute', autospec=True)
+class TriggerDeviceRescanTestCase(base.IronicLibTestCase):
+    def test_trigger(self, mock_execute):
+        self.assertTrue(disk_utils.trigger_device_rescan('/dev/fake'))
+        mock_execute.assert_has_calls([
+            mock.call('sync'),
+            mock.call('udevadm', 'settle'),
+            mock.call('partprobe', '/dev/fake', run_as_root=True, attempts=10),
+            mock.call('sgdisk', '-v', '/dev/fake', run_as_root=True),
+        ])
+
+    def test_custom_attempts(self, mock_execute):
+        self.assertTrue(
+            disk_utils.trigger_device_rescan('/dev/fake', attempts=1))
+        mock_execute.assert_has_calls([
+            mock.call('sync'),
+            mock.call('udevadm', 'settle'),
+            mock.call('partprobe', '/dev/fake', run_as_root=True, attempts=1),
+            mock.call('sgdisk', '-v', '/dev/fake', run_as_root=True),
+        ])
+
+    def test_fails(self, mock_execute):
+        mock_execute.side_effect = [('', '')] * 3 + [
+            processutils.ProcessExecutionError
+        ]
+        self.assertFalse(disk_utils.trigger_device_rescan('/dev/fake'))
+        mock_execute.assert_has_calls([
+            mock.call('sync'),
+            mock.call('udevadm', 'settle'),
+            mock.call('partprobe', '/dev/fake', run_as_root=True, attempts=10),
+            mock.call('sgdisk', '-v', '/dev/fake', run_as_root=True),
+        ])
